/*
 * Copyright (c) 2007, 2008 by Hartmut Birr
 *
 * This program is free software; you can redistribute it and/or
 * mmodify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include <inttypes.h>
#include <avr/pgmspace.h>

#include <math.h>


#include "dcg.h"
#include "encoder.h"
#include "parser.h"
#include "lcd.h"

#define NDEBUG
#include "debug.h"


int8_t EncDiff;
uint8_t Mode;
uint16_t ModifyTimer;
uint8_t DisplayTimer;
uint8_t ModifyPos;
uint8_t ModifyChr;


enum
{
    ModifyFirst = 0,
    ModifyVolt = ModifyFirst,
    ModifyAmp,
    ModifyLast = ModifyAmp
} Modify = ModifyVolt;

struct
{
    uint8_t BtnDown         : 1;    ///< state of the DOWN button
    uint8_t BtnUp           : 1;    ///< state of the UP button
    uint8_t BtnEnter        : 1;    ///< state of the ENTER button
    uint8_t lastBtnDown     : 1;    ///< previous state of the DOWN button
    uint8_t lastBtnUp       : 1;    ///< previous state of the UP button
    uint8_t lastBtnEnter    : 1;    ///< previous state of the ENTER button
    uint8_t BtnDownCnt      : 3;    ///< counter which is used to detect if the DOWN button id pressed
    uint8_t BtnUpCnt        : 3;    ///< counter which is used to detect if the UP button id pressed
    uint8_t BtnEnterCnt     : 3;    ///< counter which is used to detect if the ENTER button id pressed
} PanelData;

//---------------------------------------------------------------------------------------------
/// This function calculates the position of the visible starting digit and the position of
/// the decimal point of the given voltage.
///
/// \param U : voltage value in uV
///
/// \return uint8_t : high nibble -> position of decimal point
///                   low nibble  -> position of the first visible digit
//---------------------------------------------------------------------------------------------
uint8_t GetVoltagePosition(double U)
{
    if (U < 0.0)
    {
        U = 0.0;
    }
    if (U < 10.0)
    {
        return 0x63;
    }
    else
    {
        return 0x64;
    }
}

//---------------------------------------------------------------------------------------------
/// This function calculates the position of the visible starting digit and the position of
/// the decimal point of the given current.
///
/// \param U : current value in uA
///
/// \return uint8_t : high nibble -> position of decimal point
///                   low nibble  -> position of the first visible digit
//---------------------------------------------------------------------------------------------
uint8_t GetCurrentPosition(double I)
{
    if (I < 0.0)
    {
        I = 0.0;
    }
    if (I < 0.01)
    {
        return 0x30;
    }
    else if (I < 0.1)
    {
        return 0x31;
    }
    else if (I < 1.0)
    {
        return 0x32;
    }
    else
    {
        return 0x63;
    }
}

//---------------------------------------------------------------------------------------------
/// This function converts a voltage value into a string.
///
/// \param str   : string, which stores the result
/// \param u : voltage value, which should be convert
//---------------------------------------------------------------------------------------------
void VoltageToString(char* str, double u)
{
    if (u < 0)
    {
        u = 0;
    }

    if (u < 10)
    {
        sprintf_P(str, PSTR("%.3f V "), u);
    }
    else
    {
        sprintf_P(str, PSTR("%.2f V "), u);
    }
}

//---------------------------------------------------------------------------------------------
/// This function converts a current value into a string.
///
/// \param str   : string, which stores the result
/// \param i : current value, which should be convert
//---------------------------------------------------------------------------------------------
void CurrentToString(char* str, double i)
{
    if (i < 0.0)
    {
        i = 0.0;
    }
    if (i < 0.01)
    {
        sprintf_P(str, PSTR("%.3fmA "), i * 1e3);
    }
    else if (i < 0.1)
    {
        sprintf_P(str, PSTR("%.2fmA "), i * 1e3);
    }
    else if (i < 1.0)
    {
        sprintf_P(str, PSTR("%.1fmA "), i * 1e3);
    }
    else
    {
        sprintf_P(str, PSTR("%.3f A "), i);
    }
}

void RangeCurrentToString(char* str, uint8_t Range)
{
    uint8_t i;
    if (Range <= DC2000mA)
    {
        if (Params.MaxCurrent[Range] < 0.001)
        {
            i = sprintf_P(str, PSTR("  %.0f\344A"), Params.MaxCurrent[Range] * 1e6);
        }
        else if (Params.MaxCurrent[Range] < 1.0)
        {
            i = sprintf_P(str, PSTR("  %.0fmA"), Params.MaxCurrent[Range] * 1e3);
        }
        else if (Params.MaxCurrent[Range] < 1000.0)
        {
            i = sprintf_P(str, PSTR("  %.0fA"), Params.MaxCurrent[Range]);
        }
        else
        {
            i = sprintf_P(str, PSTR("  %.0fkA"), Params.MaxCurrent[Range] * 0.001f);
        }
    }
    else
    {
        i = sprintf_P(str, PSTR("  auto"));
    }
    while (i < 8)
    {
        str[i++] = ' ';
    }
    str[8] = 0;
}

void RangeVoltageToString(char* str, uint8_t Range)
{
    uint8_t i;
    if (Range <= 1)
    {
        i = sprintf_P(str, PSTR("  %.1fV"), Params.MaxVoltage[Range]);
    }
    else
    {
        i = sprintf_P(str, PSTR("  auto"));
    }
    while (i < 8)
    {
        str[i++] = ' ';
    }
    str[8] = 0;
}

void PanelPrintU(uint8_t cursor)
{
    char line[9];
    VoltageToString(line, xVoltage);
    line[7] = cursor;
    LCDWrite(0, 0, 8, line);
}

void PanelPrintI(uint8_t cursor)
{
    char line[9];
    double tmpCurrent;
    if (Status.FuseBlown)
    {
        sprintf_P(line, PSTR("%S"), PSTR("FUSEBLW "));
    }
    else if (Status.OverVolt)
    {
        sprintf_P(line, PSTR("%S"), PSTR("OVRVOLT "));
    }
    else if (Status.OverTemp)
    {
        sprintf_P(line, PSTR("%S"), PSTR("OVRTEMP "));
    }
    else
    {
        tmpCurrent = xCurrent < 0.0 ? 0.0 : xCurrent;
        switch (RangeI)
        {
            case DC2mA:
                if (Params.Options.ADC16Present && tmpCurrent < 0.001)
                {
                    sprintf_P(line, PSTR("%5.1f\344A "), tmpCurrent * 1e6);
                }
                else
                {
                    sprintf_P(line, PSTR("%5.3fmA "), tmpCurrent * 1e3);
                }
                break;

            case DC20mA:
                if (Params.Options.ADC16Present && tmpCurrent < 0.01)
                {
                    sprintf_P(line, PSTR("%5.3fmA "), tmpCurrent * 1e3);
                }
                else
                {
                    sprintf_P(line, PSTR("%5.2fmA "), tmpCurrent * 1e3);
                }
                break;

            case DC200mA:
                if (Params.Options.ADC16Present && tmpCurrent < 0.1)
                {
                    sprintf_P(line, PSTR("%5.2fmA "), tmpCurrent * 1e3);
                }
                else
                {
                    sprintf_P(line, PSTR("%5.1fmA "), tmpCurrent * 1e3);
                }
                break;

            case DC2000mA:
            default:
                if (Params.Options.ADC16Present && tmpCurrent < 1.0)
                {
                    sprintf_P(line, PSTR("%5.1fmA "), tmpCurrent * 1e3);
                }
                else
                {
                    sprintf_P(line, PSTR("%5.3f A "), tmpCurrent);
                }
                break;
        }
    }
    line[7] = cursor;
    LCDWrite(0, 1, 8, line);
}

double GetIncrement(uint8_t pos)
{
    return pow(10, pos - 6);
}

void jobPanel(void)
{
    if (Flags.LCDPresent)
    {
        uint8_t ButtonPressed = 0;
        uint8_t IncValid = 0;
        uint8_t Update = 0;
        uint8_t tmp;
        static char DisplayStr[9];
        static int8_t ModifyVoltagePos = -1;
        static int8_t ModifyCurrentPos = -1;
        static uint16_t BtnEnterCount = 0;
        static uint8_t Range;
        uint8_t lastModify = Modify;
        uint8_t lastMode = Mode;

        if (ModifyVoltagePos == -1)
        {
            ModifyVoltagePos = 5;
        }

        if (ModifyCurrentPos == -1)
        {
            ModifyCurrentPos = 3;
        }

        if (ModifyTimer)
        {
            ModifyTimer--;
        }

        if (DisplayTimer)
        {
            DisplayTimer--;
        }

        // save button state
        PanelData.lastBtnDown = PanelData.BtnDown;
        PanelData.lastBtnUp = PanelData.BtnUp;
        PanelData.lastBtnEnter = PanelData.BtnEnter;

        tmp = LCDGetButton();

        // button down
        if (tmp & BUTTON_DOWN)
        {
            if (PanelData.BtnDownCnt < 7)
            {
                PanelData.BtnDownCnt++;
            }
        }
        else
        {
            if (PanelData.BtnDownCnt > 0)
            {
                PanelData.BtnDownCnt--;
            }
        }
        if (PanelData.BtnDown)
        {
            if (PanelData.BtnDownCnt == 0)
            {
                PanelData.BtnDown = 0;
            }
        }
        else
        {
            if (PanelData.BtnDownCnt == 7)
            {
                PanelData.BtnDown = 1;
            }
        }

        // button up
        if (tmp & BUTTON_UP)
        {
            if (PanelData.BtnUpCnt < 7)
            {
                PanelData.BtnUpCnt++;
            }
        }
        else
        {
            if (PanelData.BtnUpCnt > 0)
            {
                PanelData.BtnUpCnt--;
            }
        }
        if (PanelData.BtnUp)
        {
            if (PanelData.BtnUpCnt == 0)
            {
                PanelData.BtnUp = 0;
            }
        }
        else
        {
            if (PanelData.BtnUpCnt == 7)
            {
                PanelData.BtnUp = 1;
            }
        }

        // button enter
        if (tmp & BUTTON_ENTER)
        {
            if (PanelData.BtnEnterCnt < 7)
            {
                PanelData.BtnEnterCnt++;
            }
        }
        else
        {
            if (PanelData.BtnEnterCnt > 0)
            {
                PanelData.BtnEnterCnt--;
            }
        }
        if (PanelData.BtnEnter)
        {
            if (PanelData.BtnEnterCnt == 0)
            {
                PanelData.BtnEnter = 0;
            }
        }
        else
        {
            if (PanelData.BtnEnterCnt == 7)
            {
                PanelData.BtnEnter = 1;
            }
        }

        if (PanelData.BtnDown && !PanelData.lastBtnDown)
        {
            DPRINT(PSTR("Down button pressed\n"));
            ButtonPressed = 1;
            if (Modify < ModifyLast)
            {
                Modify++;
            }
            Mode = 0;
            SerPrompt(NoErr, 65);
        }

        if(PanelData.BtnUp && !PanelData.lastBtnUp)
        {
            DPRINT(PSTR("Up button pressed\n"));
            ButtonPressed = 1;
            if(Modify > ModifyFirst)
            {
                Modify--;
            }
            Mode = 0;
            SerPrompt(NoErr, 66);
        }

        if (PanelData.BtnEnter)
        {
            if(!PanelData.lastBtnEnter)
            {
                DPRINT(PSTR("Enter button pressed\n"));
                ButtonPressed = 1;
                if (Mode == 2)
                {
                    Mode = 0;
                }
                else
                {
                    Mode = Mode ? 0 : 1;
                }
                SerPrompt(NoErr, 67);
            }
            if (BtnEnterCount <= 750)
            {
                BtnEnterCount++;
                ModifyTimer = 750;
            }
            if (BtnEnterCount == 750)
            {
                ButtonPressed = 1;
                Mode = 2;

            }
        }
        else
        {
            BtnEnterCount = 0;
        }

        EncDiff += GetAndResetEncPos();
        if (EncDiff <=  -Params.IncRast || EncDiff >=  Params.IncRast)
        {
            IncValid = 1;
            EncDiff /= Params.IncRast;

            if (Mode == 0)
            {
                if (EncDiff > 2)
                {
                    EncDiff *= 4;
                }
                else if (EncDiff > 1)
                {
                    EncDiff *= 2;
                }
            }
        }

        if (Mode != lastMode ||
            Modify != lastModify ||
            ButtonPressed ||
            IncValid ||
            (Mode > 0 && ModifyTimer == 0))
        {
            uint8_t StartPos;
            uint8_t PointPos;

            if ((Mode != lastMode && lastMode == 2) ||
                (ModifyTimer == 0 && lastMode == 2))
            {
                if (lastModify == ModifyVolt)
                {
                    LockRangeU = Range;
                }
                if (lastModify == ModifyAmp)
                {
                    LockRangeI = Range;
                }
                CheckLimits();
                SetLevelDAC();
            }


            if (ModifyTimer == 0 && Mode > 0)
            {
                Mode = 0;
            }

            SetActivityTimer(75);
            ModifyTimer = 750;
            Status.Busy = 1;
            Update = 1;

            switch (Modify)
            {
                case ModifyVolt:
                    if (Mode == 2)
                    {
                        if (BtnEnterCount == 750)
                        {
                            Range = LockRangeU;
                        }
                        if (IncValid)
                        {
                            if (EncDiff > 0)
                            {
                                Range++;
                            }
                            else
                            {
                                Range--;
                            }
                            EncDiff = 0;
                            if (Range == 254)
                            {
                                Range = 1;
                            }
                            else if (Range > 1)
                            {
                                Range = 255;
                            }
                        }
                        RangeVoltageToString(DisplayStr, Range);
                    }
                    else
                    {
                        tmp = GetVoltagePosition(wVoltage);
                        StartPos = tmp % 16;
                        PointPos = tmp / 16;
                        if (ModifyVoltagePos < StartPos)
                        {
                            ModifyVoltagePos = StartPos;
                        }
                        else if (ModifyVoltagePos > StartPos + 3)
                        {
                            ModifyVoltagePos = StartPos + 3;
                        }

                        if (IncValid)
                        {
                            if (Mode == 1)
                            {
                                ModifyVoltagePos -= EncDiff;
                                if (ModifyVoltagePos < StartPos)
                                {
                                    ModifyVoltagePos = StartPos;
                                }
                                else if (ModifyVoltagePos > StartPos + 3)
                                {
                                    ModifyVoltagePos = StartPos + 3;
                                }
                            }
                            else
                            {
                                wVoltage += GetIncrement(ModifyVoltagePos) * EncDiff;
                                wVoltage = floor(wVoltage * 1e3 + 0.5) * 1e-3;

                                DCVoltMod = 1;
                                CheckLimits();
                                SetLevelDAC();
                                ParseGetParam(0);

                                tmp = GetVoltagePosition(wVoltage);
                                StartPos = tmp % 16;
                                PointPos = tmp / 16;
                                if (ModifyVoltagePos < StartPos)
                                {
                                    ModifyVoltagePos = StartPos;
                                }
                                else if (ModifyVoltagePos > StartPos + 3)
                                {
                                    ModifyVoltagePos = StartPos + 3;
                                }
                            }
                            EncDiff = 0;
                        }

                        ModifyPos = 4 - (ModifyVoltagePos - StartPos);
                        if (ModifyVoltagePos >= PointPos)
                        {
                            ModifyPos--;
                        }
                        VoltageToString(DisplayStr, wVoltage);
                        ModifyChr = DisplayStr[ModifyPos];
                    }
                    break;

                case ModifyAmp:
                    if (Mode == 2)
                    {
                        if (BtnEnterCount == 750)
                        {
                            Range = LockRangeI;
                        }
                        if (IncValid)
                        {
                            if (EncDiff > 0)
                            {
                                Range++;
                            }
                            else
                            {
                                Range--;
                            }
                            EncDiff = 0;
                            if (Range == 254)
                            {
                                Range = DC2000mA;
                            }
                            else if (Range > DC2000mA)
                            {
                                Range = 255;
                            }
                        }
                        RangeCurrentToString(DisplayStr, Range);
                    }
                    else
                    {
                        tmp = GetCurrentPosition(wCurrent);
                        StartPos = tmp % 16;
                        PointPos = tmp / 16;
                        if (ModifyCurrentPos < StartPos)
                        {
                            ModifyCurrentPos = StartPos;
                        }
                        else if (ModifyCurrentPos > StartPos + 3)
                        {
                            ModifyCurrentPos = StartPos + 3;
                        }

                        if (IncValid)
                        {
                            if (Mode == 1)
                            {
                                ModifyCurrentPos -= EncDiff;
                                if (ModifyCurrentPos < StartPos)
                                {
                                    ModifyCurrentPos = StartPos;
                                }
                                else if (ModifyCurrentPos > StartPos + 3)
                                {
                                    ModifyCurrentPos = StartPos + 3;
                                }
                            }
                            else
                            {
                                wCurrent += GetIncrement(ModifyCurrentPos) * EncDiff;
                                wCurrent = floor(wCurrent * 1e6 + 0.5) * 1e-6;

                                DCAmpMod = 1;
                                CheckLimits();
                                SetLevelDAC();
                                ParseGetParam(1);

                                tmp = GetCurrentPosition(wCurrent);
                                StartPos = tmp % 16;
                                PointPos = tmp / 16;
                                if (ModifyCurrentPos < StartPos)
                                {
                                    ModifyCurrentPos = StartPos;
                                }
                                else if (ModifyCurrentPos > StartPos + 3)
                                {
                                    ModifyCurrentPos = StartPos + 3;
                                }
                            }
                            EncDiff = 0;
                        }
                        ModifyPos = 4 - (ModifyCurrentPos - StartPos);
                        if (ModifyCurrentPos >= PointPos)
                        {
                            ModifyPos--;
                        }
                        CurrentToString(DisplayStr, wCurrent);
                        ModifyChr = DisplayStr[ModifyPos];
                    }
                    break;

            }
        }

        if (DisplayTimer == 0 || DisplayTimer == 100)
        {
            Update = 1;
            if (DisplayTimer == 0)
            {
                DisplayTimer = 200;
            }
        }

        if (ModifyTimer == 0)
        {
            Mode = 0;
            Status.Busy = 0;
        }

        if (Update)
        {
            switch(Modify)
            {
                case ModifyVolt:
                    if (Mode == 2)
                    {
                        LCDWrite(0, 0, 8, "Range U ");
                        LCDWrite(0, 1, 8, DisplayStr);
                    }
                    else
                    {
                        PanelPrintI(' ');
                        if (ModifyTimer)
                        {
                            if (ModifyPos <= 4)
                            {
                                if (DisplayTimer > 100)
                                {
                                    DisplayStr[ModifyPos] = Mode ? BLOCK_HATCHED : BLOCK_SOLID;
                                }
                                else
                                {
                                    DisplayStr[ModifyPos] = ModifyChr;
                                }
                            }
                            DisplayStr[7] = Mode ? CURSOR_HATCHED : CURSOR_SOLID;
                            LCDWrite(0, 0, 8, DisplayStr);
                        }
                        else
                        {
                            PanelPrintU(CURSOR_CONCAVE);
                        }
                    }
                    break;

                case  ModifyAmp:
                    if (Mode == 2)
                    {
                        LCDWrite(0, 0, 8, "Range I ");
                        LCDWrite(0, 1, 8, DisplayStr);
                    }
                    else
                    {
                        PanelPrintU(' ');

                        if (ModifyTimer)
                        {
                            if (ModifyPos <= 4)
                            {
                                if (DisplayTimer > 100)
                                {
                                    DisplayStr[ModifyPos] = Mode ? BLOCK_HATCHED : BLOCK_SOLID;
                                }
                                else
                                {
                                    DisplayStr[ModifyPos] = ModifyChr;
                                }
                            }
                            DisplayStr[7] = Mode ? CURSOR_HATCHED : CURSOR_SOLID;
                            LCDWrite(0, 1, 8, DisplayStr);
                        }
                        else
                        {
                            PanelPrintI(CURSOR_CONCAVE);
                        }
                    }
                    break;
            }
        }
    }
}
