/*
 * Copyright (c) 2007, 2008 by Hartmut Birr
 *
 * This program is free software; you can redistribute it and/or
 * mmodify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include <inttypes.h>
#include <avr/pgmspace.h>

#include <stdio.h>

#include "dcg.h"
#include "parser.h"
#include "timer.h"

#define NDEBUG
#include "debug.h"

const MNEMONIC Mnemonics []=
{
    {"nop", 0},
    {"val", 0},
    {"dcv", 0},     // Spannung U soll in V
    {"dca", 1},     // Strom I soll in A, mA od. uA
    {"msv", 10},    // Mewert U ist in V
    {"msa", 11},    // Mewert I ist in A, mA od. uA
    {"pcv", 20},    // Prozentwert fr Ausgangsspannung
    {"pca", 21},    // Prozentwert fr Ausgangsstrom
    {"dsp", 80},    // Inkrementalgeber Impulse pro Rastpunkt
    {"all", 99},    // Mewerte
    {"ofs", 100},   // Offset DAC/ADC
    {"scl", 200},   // Skalierung DAC/ADC
    {"raw", 50},    // Rohdaten DAC/ADC
    {"opt", 150},   // Parameter/Optionen
    {"wen", 250},   // Write Enable fr EEPROM
    {"erc", 251},   // Fehlerzhler
    {"sbd", 252},   // Baudrateneinstellung
    {"idn", 254},   // Identifizierung
    {"str", 255},   // Status
    { NULL, 0}      // Terminator
};

void SendTrackCmd(void)
{
    if (TrackCh != 255)
    {
        printf_P(PSTR("%d:0=%.6f!\n"), TrackCh, wVoltage);
        printf_P(PSTR("%d:1=%.6f!\n"), TrackCh, wCurrent);
    }
}

#define PARAM_DOUBLE    0
#define PARAM_INT       1
#define PARAM_BYTE      2
#define PARAM_STR       3

#define SCALE_NONE      0
#define SCALE_A         1
#define SCALE_mA        2
#define SCALE_uA        3
#define SCALE_PROZ      4
#define SCALE_TEMP      5

typedef struct _PARAMTABLE
{
    uint8_t SubCh;
    union
    {
        double (*get_f_Function)(void);
        int32_t (*get_l_Function)(void);
        int16_t (*get_i_Function)(void);
        uint8_t (*get_b_Function)(void);
        char* (*get_s_Function)(void);
        void (*doFunction)(struct _PARAMTABLE*);
        struct
        {
            union
            {
                double* f;
                int32_t* l;
                int16_t* i;
                uint16_t* u;
                uint8_t* b;
                const char* s;
            } ram;
            union
            {
                double* f;
                int32_t* l;
                int16_t* i;
                uint16_t* u;
                uint8_t* b;
                const char* s;
            } eep;
        }s;
    }u;
    uint8_t type        : 2;
    uint8_t scale       : 3;
    uint8_t rw          : 1;
    uint8_t fct         : 2;
} PARAMTABLE;

int16_t GetADC2(void)
{
    return GetADC(2);
}

int16_t GetADC3(void)
{
    return GetADC(3);
}

int16_t GetADC4(void)
{
    return GetADC(4);
}

void GetAll(PARAMTABLE* ParamTable __attribute__((unused)))
{
    ParseGetParam(10);
    ParseGetParam(11);
    ParseGetParam(15);
}

void ReturnInput(PARAMTABLE* ParamTable __attribute__((unused)))
{
    printf_P(PSTR("%s\n"), SerInpStr);
}

PARAMTABLE SetParamTable[] PROGMEM =
{
    // Die Eintrge mssen nach SubCh sortiert sein
  {.SubCh = 0,   .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &wVoltage, .eep.f = (double*)-1}},
  {.SubCh = 1,   .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_A,    .u.s = {.ram.f = &wCurrent, .eep.f = (double*)-1}},
  {.SubCh = 2,   .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_mA,   .u.s = {.ram.f = &wCurrent, .eep.f = (double*)-1}},
  {.SubCh = 3,   .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_uA,   .u.s = {.ram.f = &wCurrent, .eep.f = (double*)-1}},
  {.SubCh = 10,  .rw = 0, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &xVoltage, .eep.f = (double*)-1}},
  {.SubCh = 11,  .rw = 0, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_A,    .u.s = {.ram.f = &xCurrent, .eep.f = (double*)-1}},
  {.SubCh = 12,  .rw = 0, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_mA,   .u.s = {.ram.f = &xCurrent, .eep.f = (double*)-1}},
  {.SubCh = 13,  .rw = 0, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_uA,   .u.s = {.ram.f = &xCurrent, .eep.f = (double*)-1}},
  {.SubCh = 15,  .rw = 0, .fct = 1, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.get_f_Function = GetPowerIn},
  {.SubCh = 20,  .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_PROZ, .u.s = {.ram.f = &DCVoltMod, .eep.f = (double*)-1}},
  {.SubCh = 21,  .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_PROZ, .u.s = {.ram.f = &DCAmpMod, .eep.f = (double*)-1}},
  {.SubCh = 50,  .rw = 0, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s.ram.u = &ADCRawU},
  {.SubCh = 51,  .rw = 0, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s.ram.u = &ADCRawI},
  {.SubCh = 52,  .rw = 0, .fct = 1, .type = PARAM_INT,    .scale = SCALE_NONE, .u.get_i_Function = GetADC2},
  {.SubCh = 53,  .rw = 0, .fct = 1, .type = PARAM_INT,    .scale = SCALE_NONE, .u.get_i_Function = GetADC3},
  {.SubCh = 54,  .rw = 0, .fct = 1, .type = PARAM_INT,    .scale = SCALE_NONE, .u.get_i_Function = GetADC4},
  {.SubCh = 70,  .rw = 0, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s.ram.u = &DACRawU},
  {.SubCh = 71,  .rw = 0, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s.ram.u = &DACRawI},
  {.SubCh = 89,  .rw = 1, .fct = 0, .type = PARAM_BYTE,   .scale = SCALE_NONE, .u.s = {.ram.b = &Params.IncRast, .eep.b = &eepParams.IncRast}},
  {.SubCh = 99,  .rw = 0, .fct = 2, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.doFunction = GetAll},
  {.SubCh = 100, .rw = 1, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s = {.ram.i = &Params.DACUOffsets[0], .eep.i = &eepParams.DACUOffsets[0]}},
  {.SubCh = 101, .rw = 1, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s = {.ram.i = &Params.DACUOffsets[1], .eep.i = &eepParams.DACUOffsets[1]}},
  {.SubCh = 102, .rw = 1, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s = {.ram.i = &Params.DACIOffsets[0], .eep.i = &eepParams.DACIOffsets[0]}},
  {.SubCh = 103, .rw = 1, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s = {.ram.i = &Params.DACIOffsets[1], .eep.i = &eepParams.DACIOffsets[1]}},
  {.SubCh = 104, .rw = 1, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s = {.ram.i = &Params.DACIOffsets[2], .eep.i = &eepParams.DACIOffsets[2]}},
  {.SubCh = 105, .rw = 1, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s = {.ram.i = &Params.DACIOffsets[3], .eep.i = &eepParams.DACIOffsets[3]}},
  {.SubCh = 110, .rw = 1, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s = {.ram.i = &Params.ADCUOffsets[0], .eep.i = &eepParams.ADCUOffsets[0]}},
  {.SubCh = 111, .rw = 1, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s = {.ram.i = &Params.ADCUOffsets[1], .eep.i = &eepParams.ADCUOffsets[1]}},
  {.SubCh = 112, .rw = 1, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s = {.ram.i = &Params.ADCIOffsets[0], .eep.i = &eepParams.ADCIOffsets[0]}},
  {.SubCh = 113, .rw = 1, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s = {.ram.i = &Params.ADCIOffsets[1], .eep.i = &eepParams.ADCIOffsets[1]}},
  {.SubCh = 114, .rw = 1, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s = {.ram.i = &Params.ADCIOffsets[2], .eep.i = &eepParams.ADCIOffsets[2]}},
  {.SubCh = 115, .rw = 1, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s = {.ram.i = &Params.ADCIOffsets[3], .eep.i = &eepParams.ADCIOffsets[3]}},
  {.SubCh = 150, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.InitVoltage, .eep.f = &eepParams.InitVoltage}},
  {.SubCh = 151, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.InitCurrent, .eep.f = &eepParams.InitCurrent}},
  {.SubCh = 152, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.GainPre, .eep.f = &eepParams.GainPre}},
  {.SubCh = 153, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.GainOut, .eep.f = &eepParams.GainOut}},
  {.SubCh = 154, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.GainI, .eep.f = &eepParams.GainI}},
  {.SubCh = 155, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.RefVoltage, .eep.f = &eepParams.RefVoltage}},
  {.SubCh = 156, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.MaxVoltage[1], .eep.f = &eepParams.MaxVoltage[1]}},
  {.SubCh = 157, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.RSense[0], .eep.f = &eepParams.RSense[0]}},
  {.SubCh = 158, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.RSense[1], .eep.f = &eepParams.RSense[1]}},
  {.SubCh = 159, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.RSense[2], .eep.f = &eepParams.RSense[2]}},
  {.SubCh = 160, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.RSense[3], .eep.f = &eepParams.RSense[3]}},
  {.SubCh = 161, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.MaxCurrent[0], .eep.f = &eepParams.MaxCurrent[0]}},
  {.SubCh = 162, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.MaxCurrent[1], .eep.f = &eepParams.MaxCurrent[1]}},
  {.SubCh = 163, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.MaxCurrent[2], .eep.f = &eepParams.MaxCurrent[2]}},
  {.SubCh = 164, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.MaxCurrent[3], .eep.f = &eepParams.MaxCurrent[3]}},
  {.SubCh = 165, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.ADCUfacs[0], .eep.f = &eepParams.ADCUfacs[0]}},
  {.SubCh = 166, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.ADCUfacs[1], .eep.f = &eepParams.ADCUfacs[1]}},
  {.SubCh = 167, .rw = 1, .fct = 0, .type = PARAM_BYTE,   .scale = SCALE_NONE, .u.s = {.ram.b = (uint8_t*)&Params.Options, .eep.b = (uint8_t*)&eepParams.Options}},
  {.SubCh = 168, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.MaxVoltage[0], .eep.f = &eepParams.MaxVoltage[0]}},
  {.SubCh = 170, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.RelayVoltage, .eep.f = &eepParams.RelayVoltage}},
  {.SubCh = 171, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.FanOnTemp, .eep.f = &eepParams.FanOnTemp}},
  {.SubCh = 200, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.DACUScales[0], .eep.f = &eepParams.DACUScales[0]}},
  {.SubCh = 201, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.DACUScales[1], .eep.f = &eepParams.DACUScales[1]}},
  {.SubCh = 202, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.DACIScales[0], .eep.f = &eepParams.DACIScales[0]}},
  {.SubCh = 203, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.DACIScales[1], .eep.f = &eepParams.DACIScales[1]}},
  {.SubCh = 204, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.DACIScales[2], .eep.f = &eepParams.DACIScales[2]}},
  {.SubCh = 205, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.DACIScales[3], .eep.f = &eepParams.DACIScales[3]}},
  {.SubCh = 210, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.ADCUScales[0], .eep.f = &eepParams.ADCUScales[0]}},
  {.SubCh = 211, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.ADCUScales[1], .eep.f = &eepParams.ADCUScales[1]}},
  {.SubCh = 212, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.ADCIScales[0], .eep.f = &eepParams.ADCIScales[0]}},
  {.SubCh = 213, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.ADCIScales[1], .eep.f = &eepParams.ADCIScales[1]}},
  {.SubCh = 214, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.ADCIScales[2], .eep.f = &eepParams.ADCIScales[2]}},
  {.SubCh = 215, .rw = 1, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_NONE, .u.s = {.ram.f = &Params.ADCIScales[3], .eep.f = &eepParams.ADCIScales[3]}},
  {.SubCh = 233, .rw = 0, .fct = 0, .type = PARAM_DOUBLE, .scale = SCALE_TEMP, .u.s = {.ram.f = &Temperature}},
  {.SubCh = 251, .rw = 1, .fct = 0, .type = PARAM_INT,    .scale = SCALE_NONE, .u.s = {.ram.u = &ErrCount, .eep.u = (uint16_t*)-1}},
  {.SubCh = 252, .rw = 1, .fct = 0, .type = PARAM_BYTE,   .scale = SCALE_NONE, .u.s = {.ram.b = &Params.SerBaudReg, .eep.b = &eepParams.SerBaudReg}},
  {.SubCh = 253, .rw = 0, .fct = 2, .type = PARAM_STR,    .scale = SCALE_NONE, .u.doFunction=ReturnInput},
  {.SubCh = 254, .rw = 0, .fct = 0, .type = PARAM_STR,    .scale = SCALE_NONE, .u.s = {.ram.s = VersStrLong}}
};

uint8_t ParseFindParamData(PARAMTABLE* Data, uint8_t SubCh)
{
    int16_t pos, min, max;

    min = 0;
    max = sizeof(SetParamTable) / sizeof(SetParamTable[0]) - 1;

    while (min <= max)
    {
        pos = (min + max) / 2;
        memcpy_P(Data, &SetParamTable[pos], sizeof(PARAMTABLE));

        if (SubCh == Data->SubCh)
        {
            // gefunden
            return 1;
        }

        if (SubCh > Data->SubCh)
        {
            min = pos + 1;
        }
        else
        {
            max = pos - 1;
        }
    }
    return 0;
}


void ParseGetParam(uint8_t SubCh)
{
    char fmt[20];
    PARAMTABLE ParamData;
    uint8_t fract_len = 4;
    union
    {
        double f;
        int32_t l;
        int16_t i;
        uint8_t b;
        const char* s;
    }Data;

    if (SubCh == 255)
    {
        SerPrompt(NoErr, Status.u8);
        return;
    }

    if (!ParseFindParamData(&ParamData, SubCh))
    {
        CHECKPOINT;
        SerPrompt(ParamErr, 0);
        return;
    }

    if (ParamData.fct == 2)
    {
        // special function
        ParamData.u.doFunction(&ParamData);
        return;
    }

    if (ParamData.fct == 1)
    {
        // get function
        switch (ParamData.type)
        {
            case PARAM_DOUBLE:
                Data.f = ParamData.u.get_f_Function();
                break;

            case PARAM_INT:
                Data.i = ParamData.u.get_i_Function();
                break;

            case PARAM_BYTE:
                Data.b = ParamData.u.get_b_Function();
                break;

            case PARAM_STR:
                Data.s = ParamData.u.get_s_Function();
                break;

        }
    }
    else
    {
        // variable
        switch(ParamData.type)
        {
            case PARAM_DOUBLE:
                Data.f = *ParamData.u.s.ram.f;
                break;

            case PARAM_INT:
                Data.i = *ParamData.u.s.ram.i;
                break;

            case PARAM_BYTE:
                Data.b = *ParamData.u.s.ram.b;
                break;

            case PARAM_STR:
                Data.s = ParamData.u.s.ram.s;
                break;
        }
    }

    // handle scaling of float variables
    if (ParamData.type == PARAM_DOUBLE)
    {
        switch (ParamData.scale)
        {
            case SCALE_PROZ:
                Data.f *= 1e2;
                fract_len = 2;
                break;

            case SCALE_A:
                fract_len = 6 - RangeI;
                break;

            case SCALE_mA:
                Data.f *= 1e3;
                fract_len = 3;
                break;

            case SCALE_uA:
                Data.f *= 1e6;
                fract_len = 0;
                break;

            case SCALE_TEMP:
                fract_len = 1;
                break;
        }
    }

    // print the parameters
    switch(ParamData.type)
    {
        case PARAM_DOUBLE:
            // create a format string, because avrgcc doesn't handle a variable length format spezifier like this "%.*f"
            sprintf_P(fmt, PSTR("#%%d:%%d=%%.%df\n"), fract_len);
            printf(fmt, SlaveCh, SubCh, Data.f);
            break;

        case PARAM_INT:
            printf_P(PSTR("#%d:%d=%d\n"), SlaveCh, SubCh, Data.i);
            break;

        case PARAM_BYTE:
            printf_P(PSTR("#%d:%d=%u\n"), SlaveCh, SubCh, Data.b);
            break;

        case PARAM_STR:
            printf_P(PSTR("#%d:%d=%s\n"), SlaveCh, SubCh, Data.s);
            break;

    }
}

void ParseSetParam(uint8_t SubCh, double Param)
{
    if (Status.Busy)
    {
        SerPrompt(BusyErr, 0);
        return;
    }
    if (SubCh == 250)
    {
        Status.EEUnlocked = 1;
    }
    else
    {
        PARAMTABLE Data;

        if (!ParseFindParamData(&Data, SubCh) || !Data.rw)
        {
            SerPrompt(ParamErr, 0);
            return;
        }

        if (Data.u.s.eep.f != (void*)-1 && !Status.EEUnlocked)
        {
            SerPrompt(LockedErr, 0);
            return;
        }

        switch(Data.type)
        {
            case PARAM_DOUBLE:
                switch(Data.scale)
                {
                    case SCALE_PROZ:
                        *Data.u.s.ram.f = Param / 1e2;
                        break;

                    case SCALE_mA:
                        *Data.u.s.ram.f = Param / 1e3;
                        break;

                    case SCALE_uA:
                        *Data.u.s.ram.f = Param / 1e6;
                        break;

                    default:
                    case SCALE_A:
                        *Data.u.s.ram.f = Param;
                        break;
                }

                break;

            case PARAM_BYTE:
                *Data.u.s.ram.b = (uint8_t)Param;
                break;

            case PARAM_INT:
                *Data.u.s.ram.i = (int16_t)Param;
                break;

        }

        if (Data.u.s.eep.f != (double*)-1)
        {
            uint8_t size;
            switch(Data.type)
            {
                case PARAM_DOUBLE:
                    size = sizeof(double);
                    break;

                case PARAM_BYTE:
                    size = sizeof(uint8_t);
                    break;

                case PARAM_INT:
                    size = sizeof(int16_t);
                    break;

                default:
                    size = 0;
            }

            eeprom_write_block(Data.u.s.ram.f, Data.u.s.eep.f, size);
            InitScales();
        }
        Status.EEUnlocked = 0;
    }

    SerPrompt(NoErr, Status.u8);

    CheckLimits();
    SetLevelDAC();
}


