/*
 * Copyright (c) 2007, 2008 by Hartmut Birr
 *
 * This program is free software; you can redistribute it and/or
 * mmodify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <util/delay.h>

#include <string.h>

#include "config.h"
#include "uart.h"
#include "parser.h"
#include "timer.h"
#include "i2creg.h"
#include "lcd.h"
#include "encoder.h"
#ifdef RUNTIME_MEASUREMENT
#include "runtime.h"
#endif
#include "dcg.h"

#define NDEBUG
#include "debug.h"

// rechnerisch korrekte 16Bit Version
PARAMS Params = {
    .DACUOffsets        = {10, 10},                         // 0.38mV 
    .DACUScales         = {1, 1},
    .DACIOffsets        = {40, 40, 40, 40},                 // 1.5mV
    .DACIScales         = {1, 1, 1, 1},
    .ADCUOffsets        = {-306, -306},                     // -R15/(R14+R15) * 2^16
    .ADCUScales         = {1.0047, 1.0047},                 // (R15+R14)/R14
    .ADCIOffsets        = {-306, -306, -306, -306},         // -R15/(R14+R15) * 2^16
    .ADCIScales         = {1.0047, 1.0047, 1.0047, 1.0047}, // (R15+R14)/R14
    .InitVoltage        = 5.0,                              // 5V
    .InitCurrent        = 0.02,                             // 20mA
    .GainPre            = 3.0,                              // Gain OP-Amp. nach ADC, (R17/R18)+1
    .GainOut            = 3.0,                              // Gain Sense-Diff-Amp., R21/R20
    .GainI              = 0.25,                             // R33/(R33+R34)
    .RefVoltage         = 2.5,
    .MaxVoltage         = {12.1, 30.},                      // Spannungsgrenzwerte der zwei U Bereiche
    .RSense             = {470.0, 47.0, 4.7, 0.47},         // Shunts der vier Bereiche
    .MaxCurrent         = {0.002, 0.02, 0.2, 2.0},          // Stromgrenzwert der vier Bereiche
    .ADCUfacs           = {2, 6},                           // (R12+R24)/R12 u. (R24+(R12||R16))/(R12||R16)
    .Options            = {.DAC16Present = 1,
                           .ADC16Present = 1},
    .RelayVoltage       = 12.5,
    .FanOnTemp          = 50,
    .IncRast            = 4,
    .SerBaudReg         = INIT_UBRR % 256,
    .TrackChSave        = 7,
    .GainPwrIn          = 19,
    .LockRangeI         = 255,
    .LockRangeU         = 255,
    .Initialised        = 0xaa55,
};

#ifdef EEP_12BIT
// rechnerisch korrekte 12Bit Version
PARAMS eepParams EEMEM = {
    .DACUOffsets        = {5, 5},                           // 2.5mV
    .DACUScales         = {1, 1},
    .DACIOffsets        = {10, 10, 10, 10},                 // 5mV
    .DACIScales         = {1, 1, 1, 1},
    .ADCUOffsets        = {0, 0},
    .ADCUScales         = {1, 1},
    .ADCIOffsets        = {0, 0, 0, 0},
    .ADCIScales         = {1, 1, 1, 1},
    .InitVoltage        = 5.0,                              // 5,0V
    .InitCurrent        = 0.02,                             // 20mA
    .GainPre            = 5.0,
    .GainOut            = 3.0,
    .GainI              = 0.5,
    .RefVoltage         = 2.048,
    .MaxVoltage         = {6.0, 20.0},
    .RSense             = {470, 47, 4.7, 0.47},
    .MaxCurrent         = {0.002, 0.02, 0.2, 2.0},
    .ADCUfacs           = {1, 5},
    .Options            = {.DAC16Present = 0,
                           .ADC16Present = 0},
    .RelayVoltage       = 22.0,
    .FanOnTemp          = 50,
    .IncRast            = 4,
    .SerBaudReg         = INIT_UBRR % 256,
    .TrackChSave        = 255,
    .GainPwrIn          = 19,
    .LockRangeI         = 255,
    .LockRangeU         = 255,
    .Initialised        = 0xaa55,
};
#endif

#ifdef EEP_16BIT
// rechnerisch korrekte 16Bit-Version fr 5A (0,5V am Shunt, R34=7k, R42=30k, R57=0,1, R53=1, R54=10, R55=100)
PARAMS eepParams EEMEM = {
    .DACUOffsets        = {10, 10},                         // 0.38mV
    .DACUScales         = {1, 1},
    .DACIOffsets        = {80, 80, 80, 80},                 // 3mV
    .DACIScales         = {1, 1, 1, 1},
    .ADCUOffsets        = {-306, -306},                     // -R15/(R14+R15) * 2^16
    .ADCUScales         = {1.0047, 1.0047},                 // (R15+R14)/R14
    .ADCIOffsets        = {-306, -306, -306, -306},         // -R15/(R14+R15) * 2^16
    .ADCIScales         = {2.0094, 2.0094, 2.0094, 2.0094}, // (R15+R14)/R14 * (R42+R50)/R50/2
    .InitVoltage        = 5.0,                              // 5V
    .InitCurrent        = 0.02,                             // 20mA
    .GainPre            = 3.0,                              // Gain OP-Amp. nach ADC, (R17/R18)+1
    .GainOut            = 3.0,                              // Gain Sense-Diff-Amp., R21/R20
    .GainI              = 0.125,                            // R33/(R33+R34)
    .RefVoltage         = 2.5,
    .MaxVoltage         = {12.1, 30.0},                     // Spannungsgrenzwerte der zwei Bereiche
    .RSense             = {100.0, 10.0, 1, 0.1},            // Shunts der vier Bereiche
    .MaxCurrent         = {0.005, 0,05, 0.5, 5.0},          // Stromgrenzwert der vier Bereiche
    .ADCUfacs           = {2, 6},                           // (R12+R24)/R12 u. (R24+(R12||R16))/(R12||R16)
    .Options            = {.DAC16Present = 1,
                           .ADC16Present = 1},
    .RelayVoltage       = 12.5,
    .FanOnTemp          = 50,
    .IncRast            = 4,
    .SerBaudReg         = INIT_UBRR % 256,
    .TrackChSave        = 7,
    .GainPwrIn          = 19,
    .LockRangeI         = 255,                              // auto 
    .LockRangeU         = 255,                              // auto
    .Initialised        = 0xaa55,
};
#endif

#ifdef EEP_16BIT_TEST
// rechnerisch korrekte Testversion mit 16Bit-DAC, 2,5V Ref. und ADC vom ATMega
PARAMS eepParams EEMEM = {
    .DACUOffsets        = {10, 10},                         // 0.38mV
    .DACUScales         = {1, 1},
    .DACIOffsets        = {40, 40, 40, 40},                 // 1.5mV
    .DACIScales         = {1, 1, 1, 1},
    .ADCUOffsets        = {0, 0},
    .ADCUScales         = {1, 1},
    .ADCIOffsets        = {0, 0, 0, 0},
    .ADCIScales         = {1, 1, 1, 1},
    .InitVoltage        = 5.0,                              // 5V
    .InitCurrent        = 0.02,                             // 20mA
    .GainPre            = 3.0,                              // Gain OP-Amp. nach ADC, (R17/R18)+1
    .GainOut            = 3.0,                              // Gain Sense-Diff-Amp., R21/R20
    .GainI              = 0.25,                             // R33/(R33+R34)
    .RefVoltage         = 2.5,
    .MaxVoltage         = {12.1, 30.0},                     // Spannungsgrenzwerte der zwei Bereiche
    .RSense             = {470.0, 47.0, 4.7, 0.47},         // Shunts der vier Bereiche
    .MaxCurrent         = {0.002, 0.02, 0.2, 2.0},          // Stromgrenzwert der vier Bereiche
    .ADCUfacs           = {2, 6},                           // (R12+R24)/R12 u. (R24+(R12||R16))/(R12||R16)
    .Options            = {.DAC16Present = 1,
                           .ADC16Present = 0},
    .RelayVoltage       = 12.5,
    .FanOnTemp          = 50,
    .IncRast            = 4,
    .SerBaudReg         = INIT_UBRR % 256,
    .TrackChSave        = 7,
    .GainPwrIn          = 19,
    .LockRangeI         = 255,                              // auto
    .LockRangeU         = 255,                              // auto
    .Initialised        = 0xaa55,
};
#endif

#ifdef EEP_16BIT_CM
// 16Bit Version nach CM
PARAMS eepParams EEMEM = {
    .DACUOffsets        = {10, 10},                         // 0.38mV
    .DACUScales         = {1.001, 1.0032},
    .DACIOffsets        = {10, 10, 10, 10},                 // 0.38mV 
    .DACIScales         = {1.003, 1.003, 1.003, 1.003},
    .ADCUOffsets        = {-180, -180},
    .ADCUScales         = {1.005, 1.0005},
    .ADCIOffsets        = {-180, -180, -180, -180},
    .ADCIScales         = {1, 1, 1, 1},
    .InitVoltage        = 5.0,                              // 5,0V
    .InitCurrent        = 0.02,                             // 20mA
    .GainPre            = 3.0,
    .GainOut            = 3.0,
    .GainI              = 0.25,
    .RefVoltage         = 2.5,
    .MaxVoltage         = {12.1, 30.0},
    .RSense             = {470, 47, 4.7, 0.47},
    .MaxCurrent         = {0.002, 0.02, 0.2, 2.0},
    .ADCUfacs           = {2, 6},
    .Options            = {.DAC16Present = 1,
                           .ADC16Present = 1},
    .RelayVoltage       = 12.5,
    .FanOnTemp          = 50,
    .IncRast            = 4,
    .SerBaudReg         = INIT_UBRR % 256,
    .TrackChSave        = 255,
    .GainPwrIn          = 19,
    .LockRangeI         = 255,                              // auto
    .LockRangeU         = 255,                              // auto
    .Initialised        = 0xaa55,
};
#endif

#ifdef EEP_12BIT_CM
// 12Bit Version nach CM
PARAMS eepParams EEMEM = {
    .DACUOffsets        = {5, 5},                           // 2.5mV
    .DACUScales         = {1.003, 1.005},
    .DACIOffsets        = {5, 5, 5, 5},                     // 2.5mV
    .DACIScales         = {0.995, 0.991, 0.986, 0.937},
    .ADCUOffsets        = {3, 3},
    .ADCUScales         = {1, 1},
    .ADCIOffsets        = {3, 3, 3, 3},
    .ADCIScales         = {1.003, 1.008, 1.014, 1.07},
    .InitVoltage        = 5.0,                              // 5,0V
    .InitCurrent        = 0.02,                             // 20mA
    .GainPre            = 5.0,
    .GainOut            = 3.0,
    .GainI              = 0.5,
    .RefVoltage         = 2.048,
    .MaxVoltage         = {6.0, 20.0},
    .RSense             = {470, 47, 4.7, 0.47},
    .MaxCurrent         = {0.002, 0.02, 0.2, 1.0},
    .ADCUfacs           = {1, 5},
    .Options            = {.DAC16Present = 0,
                           .ADC16Present = 0},
    .RelayVoltage       = 22.0,
    .FanOnTemp          = 50,
    .IncRast            = 4,
    .SerBaudReg         = INIT_UBRR % 256,
    .TrackChSave        = 255,
    .GainPwrIn          = 19,
    .LockRangeI         = 255,                              // auto
    .LockRangeU         = 255,                              // auto
    .Initialised        = 0xaa55,
};
#endif

const char VersStrLong[]  = "0.1 [DCG2 by Hartmut Birr 11/2007]";
const prog_char VersStrShort[] = "DCG2 0.1";

uint8_t SlaveCh;
uint8_t TrackCh;
uint16_t ErrCount;
uint8_t ActivityTimer;
uint8_t RangeI;
uint8_t RangeU;
uint8_t LockRangeU;
uint8_t LockRangeI;
uint8_t InitLockRangeU;
uint8_t InitLockRangeI;

STATUS Status;
FLAGS Flags;

uint16_t ADCRawU;
uint16_t ADCRawI;

uint16_t DACRawU;
uint16_t DACRawI;
uint16_t DACMax;

double Temperature;
double DACLSBU[2];
double ADCLSBU[2];
double DACLSBI[4];
double ADCLSBI[4];
double PwrInFac;
double xVoltage;
double xCurrent;

double xMeanVoltage;
double DCVoltMod;
double DCAmpMod;

double wVoltage;
double wCurrent;

int uart_putchar(char c, FILE* stream __attribute__((unused)))
{
    if (c == '\n')
    {
        c = '\r';
        while (1 != setUartTxData((uint8_t*)&c, 1, 0));
        c = '\n';
    }
    while(1 != setUartTxData((uint8_t*)&c, 1, 0));
    return 0;
}

static FILE mystdout = FDEV_SETUP_STREAM(uart_putchar, NULL, _FDEV_SETUP_WRITE);

void CheckLimits(void)
{
    DPRINT(PSTR("wVoltage=%.5f wCurrent=%.5f\n"), wVoltage, wCurrent);

    if (wVoltage < 0)
    {
        wVoltage = 0;
    }
    else
    {
        if (LockRangeU == 255)
        {
            if (wVoltage > Params.MaxVoltage[1])
            {
                wVoltage = Params.MaxVoltage[1];
            }
        }
        else
        {
            if (wVoltage > Params.MaxVoltage[LockRangeU])
            {
                wVoltage = Params.MaxVoltage[LockRangeU];
            }
        }
    }
    if (wCurrent < 0.0)
    {
        wCurrent = 0.0;
    }
    else
    {
        if (LockRangeI == 255)
        {
            if (wCurrent > Params.MaxCurrent[DC2000mA])
            {
                wCurrent = Params.MaxCurrent[DC2000mA];
            }
        }
        else
        {
            if (wCurrent > Params.MaxCurrent[LockRangeI])
            {
                wCurrent = Params.MaxCurrent[LockRangeI];
            }
        }
    }
    if (TrackCh > 128)
    {
        TrackCh = 255;
    }
    else if (TrackCh > 7)
    {
        TrackCh = 7;
    }

    if (DCVoltMod < 0)
    {
        DCVoltMod = 0;
    }
    else if (DCVoltMod > 1)
    {
        DCVoltMod = 1;
    }

    if (DCAmpMod < 0)
    {
        DCAmpMod = 0;
    }
    else if (DCAmpMod > 1)
    {
        DCAmpMod = 1;
    }
}

void jobGetValues(void)
{
    static uint8_t lastRangeI = 0xff;
    static uint8_t lastRangeU = 0xff;
    static uint8_t waitITimer = 0;
    static uint8_t waitUTimer = 0;
    uint8_t sreg;
    union
    {
        int32_t i32;
        uint16_t u16[2];
    } Values;

    if (lastRangeI != RangeI)
    {
        lastRangeI = RangeI;
        waitITimer = 10;    // 40ms
    }

    if (lastRangeU != RangeU)
    {
        lastRangeU = RangeU;
        waitUTimer = 10;    // 40ms
    }

    if (waitUTimer)
    {
        waitUTimer--;
    }
    else
    {
        Values.u16[1] = 0;
        if (Params.Options.ADC16Present)
        {
            sreg = SREG;
            cli();
            Values.u16[0] = ADCRawU;
            SREG = sreg;
        }
        else
        {
            Values.u16[0] = GetADC(2);
        }
        xVoltage = (Values.i32 + Params.ADCUOffsets[lastRangeU]) * ADCLSBU[lastRangeU];
    }

    if (waitITimer)
    {
        waitITimer--;
    }
    else
    {
        Values.u16[1] = 0;

        if (Params.Options.ADC16Present)
        {
            sreg = SREG;
            cli();
            Values.u16[0] = ADCRawI;
            SREG = sreg;
        }
        else
        {
            Values.u16[0] = GetADC(3);
        }
        xCurrent = (Values.i32 + Params.ADCIOffsets[lastRangeI]) * ADCLSBI[lastRangeI];
    }
    xMeanVoltage = (3 * xMeanVoltage + xVoltage) / 4;
}


double GetPowerIn(void)
{
    return GetADC(4) * PwrInFac;
}

void InitScales(void)
{
    double Ufac;
    uint32_t tmpDACMax, tmpADCMax;
    uint8_t i;

    if (Params.Options.DAC16Present)
    {
        Ufac = 2 * Params.RefVoltage;   // LTC1655 intern 2fache Verstrkung
        tmpDACMax = 0x10000;
    }
    else
    {
        Ufac = Params.RefVoltage;
        tmpDACMax = 0x1000;
    }
    if (Params.Options.ADC16Present)
    {
        tmpADCMax = 0x10000;
    }
    else
    {
        tmpADCMax = 0x400;
    }
    DACLSBU[0] = Ufac * Params.GainOut / (tmpDACMax * Params.DACUScales[0]);
    DACLSBU[1] = Ufac * Params.GainPre * Params.GainOut / (tmpDACMax * Params.DACUScales[1]);
    for (i = 0; i < 2; i++)
    {
        ADCLSBU[i] = Params.ADCUfacs[i] * Params.ADCUScales[i] * Params.RefVoltage * Params.GainOut / tmpADCMax;
    }

    Ufac *= Params.GainI;   // Spannungsteiler nach DAC-I auf 1/4 (0.25), R34, R33
    for (i = 0; i < 4; i++)
    {
        // mA pro LSBit fr 4 Bereiche, Output:
        DACLSBI[i] = Ufac / Params.RSense[i] / (tmpDACMax * Params.DACIScales[i]);
        // mA pro LSBit fr 4 Bereiche, Input:
        ADCLSBI[i] = Params.ADCIScales[i] * Params.RefVoltage / (2 * Params.RSense[i]) / tmpADCMax; // X2 durch U11
    }
    PwrInFac = Params.RefVoltage * Params.GainPwrIn / 1024;
    DACMax = tmpDACMax - 1; // fr SetLevelDAC
    TrackCh = Params.TrackChSave;
    ConfigLM75(Params.FanOnTemp);

    InitLockRangeU = Params.LockRangeU;
    if (InitLockRangeU > 1 && InitLockRangeU != 255)
    {
        InitLockRangeU = 255;
    }
    InitLockRangeI = Params.LockRangeI;
    if (InitLockRangeI > DC2000mA && InitLockRangeI != 255)
    {
        InitLockRangeI = 255;
    }
}

uint8_t CalcRangeI(double I)
{
    uint8_t Range = DC2mA;

    while (Range < DC2000mA && I > Params.MaxCurrent[Range])
    {
        Range++;
    }

    return Range;
}

void SetLevelDAC(void)
{
    int32_t tmpDAC;
    uint8_t Range;
    uint8_t sreg;

    // Berechnung des Range-Wertes fr Strom
    if (LockRangeI == 255)
    {
        Range = CalcRangeI(wCurrent);
    }
    else
    {
        Range = LockRangeI;
    }

    if (RangeI != Range)
    {
        DCAmpMod = 1; // Prozent-Faktor rcksetzen
        DPRINT(PSTR("Shunt: %s\n"), Range == DC2mA ? "470R" : Range == DC20mA ? "47R" : Range == DC200mA ? "4R7" : "0R47");
    }

    // Berechnung des DAC-Wertes fr Strom
    tmpDAC = (int32_t)(wCurrent * DCAmpMod / DACLSBI[Range] + 0.5) + Params.DACIOffsets[Range];
    if (tmpDAC > DACMax)
    {
        tmpDAC = DACMax;
    }
    else if (tmpDAC < 0)
    {
        tmpDAC = 0;
    }
    sreg = SREG;
    cli();
    DACRawI = tmpDAC;
    RangeI = Range;
    SREG = sreg;

    if (LockRangeU == 255)
    {
        Range = 0;
        if (wVoltage > Params.MaxVoltage[0])
        {
            Range = 1;
        }
    }
    else
    {
        Range = LockRangeU;
    }
    if (Range != RangeU)
    {
        DCVoltMod = 1; // Prozent-Faktor rcksetzen
    }
    // Berechnung des DAC-Wertes fr Spannung
    tmpDAC = (int32_t)(wVoltage * DCVoltMod / DACLSBU[Range] + 0.5) + Params.DACUOffsets[Range];
    if (tmpDAC > DACMax)
    {
        tmpDAC = DACMax;
    }
    else if (tmpDAC < 0)
    {
        tmpDAC = 0;
    }
    sreg = SREG;
    cli();
    DACRawU = tmpDAC;
    RangeU = Range;
    SREG = sreg;
}

void jobFaultCheck(void)
{
    double tmpVolt;
    static uint8_t count = 0;

    if (++count >= 0)   // 20*100ms = 2s
    {
        Temperature = GetLM75Temp();
        if (Temperature > 80)
        {
            if (!Status.OverTemp)
            {
                DPRINT(PSTR("%-10lu LM75: %.1fC\n"), GetTicker(), Temperature);
                SerPrompt(OvlErr, Status.u8);
            }
            Status.OverTemp = 1;
        }
        else if (Temperature < 70)
        {
            if (Status.OverTemp)
            {
                Flags.PwrInRange = 0;
            }
            Status.OverTemp = 0;
        }
    }
    // Eingangsspannung berechnen
    tmpVolt = GetPowerIn() - 2;
    if (xMeanVoltage > tmpVolt)
    {
        if (!Status.OverVolt)
        {
            DPRINT(PSTR("%-10lu Vdc: %.2f, Vin: %.2f\n"), GetTicker(), xMeanVoltage, tmpVolt);
            SerPrompt(FaultErr, Status.u8);
        }
        Status.OverVolt = 1;
    }
    else
    {
        if (Status.OverVolt)
        {
            Flags.PwrInRange = 0;
        }
        Status.OverVolt = 0;
    }
    if (Flags.RelaisState == 2)
    {
        // nur Prfen, wenn Relais eingeschaltet sind
        Status.FuseBlown = (tmpVolt < 5.0);
    }
    else
    {
        Status.FuseBlown = 0;
    }
}

void jobSwitchRelay(void)
{
    static uint8_t RelayTimer = 0;
    static uint8_t CurrentModeCounter = 0;

    if (PIND & (1<<PD4))
    {
        if (CurrentModeCounter > 0)
        {
            CurrentModeCounter--;
        }
    }
    else
    {
        if (CurrentModeCounter < 30)
        {
            CurrentModeCounter++;
        }
    }
    if (Status.CurrentMode && CurrentModeCounter < 10)
    {
        Status.CurrentMode = 0;
        PORTD &= ~(1<<PD3);     // voltage LED on, current LED off
    }
    if (!Status.CurrentMode && CurrentModeCounter > 20)
    {
        Status.CurrentMode = 1;
        PORTD |= (1<<PD3);      // voltage LED off, current LED on
    }

    if (wVoltage > Params.RelayVoltage && !Status.CurrentMode)
    {
        Flags.PwrInRange = 1;
    }
    if (xMeanVoltage > Params.RelayVoltage)
    {
        Flags.PwrInRange = 1;
    }
    else if (xMeanVoltage + 0.5 < Params.RelayVoltage)   // 0.5V hysteresis
    {
        Flags.PwrInRange = 0;
    }

    if (RelayTimer)
    {
        RelayTimer--;
    }

    switch(Flags.RelaisState)
    {
        case 0: // off
            if (!Status.OverVolt && !Status.OverTemp)
            {
                if(Flags.PwrInRange)
                {
                    DPRINT(PSTR("%-10lu Relais 1: on\n"), GetTicker());
                    PORTB |= (1<<PB2);  // Relais 1 ein
                }
                else
                {
                    DPRINT(PSTR("%-10lu Relais 2: on\n"), GetTicker());
                    PORTB |= (1<<PB3);  // Relais 2 ein
                }
                RelayTimer = 2; // 8ms
                Flags.RelaisState = 1;
            }
            break;


        case 1: // Wait after switching
            if (RelayTimer == 0)
            {
                if (PORTB & ((1<<PB3)|(1<<PB2)))
                {
                    // at least one relais is on
                    Flags.RelaisState = 2;
                }
                else
                {
                    // both relais are off
                    if (Status.OverVolt || Status.OverTemp)
                    {
                        Flags.RelaisState = 0;
                    }
                    else
                    {
                        if(Flags.PwrInRange)
                        {
                            DPRINT(PSTR("%-10lu Relais 1: on\n"), GetTicker());
                            PORTB |= (1<<PB2);  // Relais 1 ein
                        }
                        else
                        {
                            DPRINT(PSTR("%-10lu Relais 2: on\n"), GetTicker());
                            PORTB |= (1<<PB3);  // Relais 2 ein
                        }
                        RelayTimer = 2; // 8ms
                    }
                }
            }
            break;

        case 2: // on
            if (Status.OverVolt || Status.OverTemp)
            {
#ifdef DEBUG
                if ((PORTB & ((1<<PB3)|(1<<PB2))) == ((1<<PB3)|(1<<PB2)))
                {
                    DPRINT(PSTR("%-10lu Relais 1,2: off\n"), GetTicker());
                }
                else if (PORTB & (1<<PB2))
                {
                    DPRINT(PSTR("%-10lu Relais 1: off\n"), GetTicker());
                }
                else if(PORTB & (1<<PB3))
                {
                    DPRINT(PSTR("%-10lu Relais 2: off\n"), GetTicker());
                }
#endif
                PORTB &= ~(1<<PB2); // Relais 1 aus
                PORTB &= ~(1<<PB3); // Relais 2 aus
                RelayTimer = 2; // 8ms
                Flags.RelaisState = 1;
            }
            else if (Flags.PwrInRange)
            {
                if (PORTB & (1<<PB3))
                {
                    // Relais 2 ist ein
                    DPRINT(PSTR("%-10lu Relais 2: off\n"), GetTicker());
                    PORTB &= ~(1<<PB3); // Relais 2 aus
                    RelayTimer = 2;
                    Flags.RelaisState = 1;
                }

            }
            else
            {
                if (PORTB & (1<<PB2))
                {
                    // Relais 1 ist ein
                    DPRINT(PSTR("%-10lu Relais 1: off\n"), GetTicker());
                    PORTB &= ~(1<<PB2); // Relais 1 aus
                    RelayTimer = 2; // 8ms
                    Flags.RelaisState = 1;
                }
            }
            break;
    }
}

void SetActivityTimer(uint8_t Value)
{
    if (Value > ActivityTimer)
    {
        ActivityTimer = Value;
        PORTD &= ~(1<<PD2);             // Activity LED an
    }
}

void jobActivityTimer(void)
{
    if (ActivityTimer)
    {
        ActivityTimer--;
        if (ActivityTimer == 0)
        {
            PORTD |= (1<<PD2);          // Activity LED aus
        }
    }
}

int __attribute__((noreturn)) main(void)
{
    uint8_t i;
    uint8_t StartTimer = 0;

    // Ports initialisieren
    DDRA = 0;
    PORTA = (1<<PA1)|(1<<PA0);

    DDRB = (1<<DDB7)|(1<<DDB5)|(1<<DDB4)|(1<<DDB3)|(1<<DDB2)|(1<<DDB1)|(1<<DDB0);
    PORTB = (1<<PB7)|(1<<PB6)|(1<<PB4)|(1<<PB1)|(1<<PB0);

    DDRC = (1<<DDC7)|(1<<DDC6)|(1<<DDC5)|(1<<DDC4)|(1<<DDC3)|(1<<DDC2);
    #ifdef DUAL_DAC
	PORTC = (1<<PC5)|(1<<PC4)|(1<<PC3)|(1<<PC2)|(1<<PC1)|(1<<PC0);
	#else
	PORTC = (1<<PC3)|(1<<PC2)|(1<<PC1)|(1<<PC0);
	#endif

    DDRD = (1<<DDD3)|(1<<DDD2);
    PORTD = (1<<PD7)|(1<<PD6)|(1<<PD5)|(1<<PD2);

    // Interrupts freigeben
    sei();

    // Timer initialisieren
    InitTimer();

    // Warten bis C's vom MAX232 geladen sind
    wait_us(20000);

    SlaveCh = ((uint8_t)~PIND) >> 5;

    // Parameter initialisieren
    if (0xaa55 == eeprom_read_word(&eepParams.Initialised))
    {
        eeprom_read_block(&Params, &eepParams, sizeof(Params));
    }
    else
    {
        eeprom_write_block(&Params, &eepParams, sizeof(eepParams));
    }

    // UART initialisieren
    initUartUBRR(Params.SerBaudReg);

    // printf auf UART verbiegen
    stdout = &mystdout;

    // I2C initialisieren
    initI2C();

    // Init LCD
    if (LCDInit())
    {
        char s[9];

        Flags.LCDPresent = 1;

        LCDWrite_P(0, 0, strlen(VersStrShort), VersStrShort);
        sprintf_P(s, PSTR("Adr  %-3d"), SlaveCh);

        LCDWrite(0, 1, 8, s);
    }

    PORTD &= ~(1<<PD2);         // Activity LED an

    wait_us(1000000);           // 1sec warten

    for (i = 0; i < SlaveCh; i++)
    {
        PORTD |= (1<<PD2);      // Activity LED aus
        wait_us(150000);
        PORTD &= ~(1<<PD2);     // Activity LED ein
        wait_us(150000);
    }
    PORTD |= (1<<PD2);          // Activity LED aus

    Status.u8 = 0;

    InitScales();

    LockRangeU = InitLockRangeU;    
    LockRangeI = InitLockRangeI;

    DCVoltMod = 1;              // Faktor, Parameter aber in Prozent
    DCAmpMod = 1;
    wVoltage = 0;

    wCurrent = Params.InitCurrent;

    SetLevelDAC();              // Anfangswerte

    wait_us(4000);              // warten, da die ersten DAC Werte ausgegeben wurden

    ParseGetParam(254);
    PORTC |= (1<<PC7);          // Outputenable

#ifdef RUNTIME_MEASUREMENT
    Runtime();
#endif

    // Timer Starten
    StartTimers();

    set_sleep_mode(SLEEP_MODE_IDLE);

    while(1)
    {
        if (TestAndResetTimerOV(TIMER_100MS))
        {
            // Funktionen mit 100ms Periode
            jobFaultCheck();
        }
        if (TestAndResetTimerOV(TIMER_50MS))
        {
            // Funktionen mit 50ms Periode
        }
        if (TestAndResetTimerOV(TIMER_10MS))
        {
            // Funktionen mit 10ms Periode
        }
        if (TestAndResetTimerOV(TIMER_4MS))
        {
            jobGetValues();

            // Funktionen mit 4ms Periode
            jobPanel();

            if (StartTimer >= 20)
            {
                jobParseData();
            }

            if (StartTimer < 255)
            {
                if (StartTimer == 4)
                {
                    wVoltage = Params.InitVoltage;
                    SetLevelDAC();              // Anfangswerte
                    Flags.PwrInRange = 0;
                    ErrCount = 0;
                }
                else if (StartTimer == 40)
                {
                    jobFaultCheck();
                    if (Status.FuseBlown)
                    {
                        SerPrompt(FuseErr,Status.u8);
                    }
                }
                StartTimer++;
            }

            jobSwitchRelay();
            jobActivityTimer();
        }
        sleep_enable();
        sleep_cpu();
        sleep_disable();
    }
}

