/*
 * Copyright (c) 2007 by Hartmut Birr
 *
 * This program is free software; you can redistribute it and/or
 * mmodify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. 
 *
 */

#include <inttypes.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "parser.h"
#include "uart.h"

#define NDEBUG
#include "debug.h"

const char* ErrStr[10] = 
{
    "[OK]",
    "[SRQUSR]",
    "[BUSY]",
    "[OVERLD]",
    "[CMDERR]",
    "[PARERR]",
    "[LOCKED]",
    "[CHKSUM]",
    "[FUSE]",
    "[FAULT]"
};

char SerInpStr[64];
uint8_t SerInpCount;

uint8_t CurrentCh = 255;
struct
{
    uint8_t Verbose : 1;
} ParserFlags;

void SerPrompt(ERROR Err, uint8_t Status)
{
    if (ParserFlags.Verbose || Err != NoErr)
    {
        printf_P(PSTR("#%d:255=%d %s\n"), SlaveCh, Err + Status, ErrStr[Err]);
    }
}

void SerStr (char* str)
{
    printf_P(PSTR("%s\n"), str);
}

void jobParseData(void)
{
    uint8_t count;
    static char c = 0;
    ERROR Error = NoErr;
    uint8_t SubCh;
    char prev;

    count = getUartRxCount();

    if (count > 20)
    {
        count = 20;
    }

    while (count--)
    {
        prev = c;
        getUartRxData((uint8_t*)&c, 1);
    
        if (c >= ' ' && (uint8_t)c <= 127)
        {
            if (SerInpCount < sizeof(SerInpStr))
            {
                SerInpStr[SerInpCount++] = c;
            }
        }
        else if (c == '\b')
        {
            if (SerInpCount)
            {
                SerInpCount--;
            }
        }
        else if (c == '\n' && prev == '\r')
        {
            continue;
        } 
        else if (c == '\r' || c == '\n')
        {
            char* s;
            char* pos;
            uint16_t value;
            double Param;

            if (SerInpCount >= sizeof(SerInpStr))
            {
                // buffer overrun
                SerInpStr[sizeof(SerInpStr) - 1] = 0;
                Error = SyntaxErr;
            }
            else
            {
                SerInpStr[SerInpCount] = 0;
            }
             
            count = 0;	            // stop while loop
            SerInpCount = 0;
            s = SerInpStr;
            while(*s == ' ') s++;
            if (*s == 0)
            {
                CHECKPOINT;
                if (Error == NoErr)
                {
                    // empty string
                    SerPrompt(NoErr, 0);
                }
                break;
            }
            if (*s == '#')
            {
                SerStr(SerInpStr);  // forward the result
                Error = NoErr;      // don't report an error for a shortened result
                break;
            }
            if (Error != NoErr)
            {
                CHECKPOINT;
                break;
            }
            pos = strchr(s, ':');
            if (pos)
            {
                if (*s == '*')
                {
                    // Omni-Befehl, Weiterleiten
                    SerStr(SerInpStr);
                    s++;
                    while(*s == ' ') s++;
                    if (s != pos)
                    {
                        CHECKPOINT;
                        Error = SyntaxErr;
                        break;
                    }
                }
                else
                {
                    value = 0;
                    if (*s >= '0' && *s <= '9')
                    {
                        value += *s - '0';
                        s++;
                    }
                    else
                    {
                        CHECKPOINT;
                        Error = SyntaxErr;
                        break;
                    }
                    while(*s == ' ') s++;
                    if (s != pos)
                    {
                        CHECKPOINT;
                        Error = SyntaxErr;
                        break;
                    }
                    CurrentCh = value;
                    if (CurrentCh != SlaveCh)
                    {
                        // Nicht fr uns, weiterleiten
                        SerStr(SerInpStr);
                        break;
                    }
                }
                s = pos + 1;
                while(*s == ' ') s++;
                if (*s == 0)
                {
                    CHECKPOINT;
                    Error = SyntaxErr;
                    break;
                }
            }
            // Ist fr uns, ab hier eigentliche Verarbeitung
            ParserFlags.Verbose = strchr(s, '!') != NULL || strchr(s, '?') != NULL;
            pos = strchr(SerInpStr, '$');
            if (pos && isxdigit(pos[1]))
            {
                uint8_t Sum = 0;
                uint8_t calcSum = 0;
                uint8_t i;
                char* ss = SerInpStr;

                while (ss < pos)
                {
                    Sum ^= *ss++;
                }
                i = isxdigit(pos[2]) ? 2 : 1;
                while (i)
                {
                    i--;
                    pos++;
                    calcSum *= 16;
                    if (isdigit(*pos)) 
                    {
                        calcSum += *pos - '0';
                    }
                    else
                    {
                        calcSum += toupper(*pos) - 'A' + 10;
                    }
                }
                DPRINT(PSTR("Sum: %02x, calculated: %02x\n"), Sum, calcSum);  
                if (Sum != calcSum)
                {
                    ErrCount++;
                    Error = ChecksumErr;
                    break;
                }
              
            }
            if (*s >= '0' && *s <= '9')
            {
                // Direkter SubCh-Aufrud
                value = 0;
                while (*s >= '0' && *s <= '9' && value <= 255)
                {
                    value *= 10;
                    value += *s - '0';
                    s++;
                }
                if (value > 255)
                {
                    Error = SyntaxErr;
                    break;
                }
                SubCh = value;
            }
            else
            {
                // Klartext bersetzen
                char cmd[4];
                uint8_t i;

                cmd[3] = 0;
                for (i = 0; i < 3; i++, s++)
                {
                    if (*s >= 'A' && *s <= 'Z')
                    {
                        cmd[i] = *s + 'a' - 'A';
                    }
                    else if (*s >= 'a' && *s <= 'z')
                    {
                        cmd[i] = *s;
                    }
                    else
                    {
                        break;
                    }
                }
                if (i < 3)
                {
                    CHECKPOINT;
                    Error = SyntaxErr;
                    break;
                }
                value = 0;
                while(*s == ' ') s++;
                if (*s >= '0' && *s <= '9')
                {
                    while (*s >= '0' && *s <= '9' && value <= 255)
                    {
                        value *= 10;
                        value += *s - '0';
                        s++;
                    }
                    if (value > 255)
                    {
                        CHECKPOINT;
                        Error = SyntaxErr;
                        break;
                    }

                }
                for (i = 0; Mnemonics[i].Mnemonic; i++)
                {
                    const char* ptr = Mnemonics[i].Mnemonic;
                    if (ptr[0] == cmd[0] &&
                        ptr[1] == cmd[1] &&
                        ptr[2] == cmd[2])
                    {
                        break;
                    }
                }
                if (!Mnemonics[i].Mnemonic)
                {
                    CHECKPOINT;
                    Error = SyntaxErr;
                    break;
                }
                if (i == 0)
                {
                    // nop
                    CHECKPOINT;
                    SerPrompt(NoErr, 0);
                    break;
                }
                if (value + Mnemonics[i].Offset > 255)
                {
                    CHECKPOINT;
                    Error = SyntaxErr;
                    break;
                }
                SubCh = value + Mnemonics[i].Offset;
            }
            while (*s == ' ') s++;
            if (*s == '=')
            {
                s++;
                while(*s == ' ') s++;
                if (!(*s >= '0' && *s <= '9') && *s != '-')
                {
                    CHECKPOINT;
                    Error = ParamErr;
                    break;
                }
                Param = atof(s);
                ParseSetParam(SubCh, Param);
            }
            else
            {
                ParseGetParam(SubCh);
            }
            SetActivityTimer(50);
        }
    }
    if (Error != NoErr)
    {
        SerPrompt(Error, 0);
    }
}

