/*
 * Copyright (c) 2007 by Hartmut Birr
 *
 * This program is free software; you can redistribute it and/or
 * mmodify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "timer.h"
#include "dcg.h"
#include "encoder.h"

static volatile struct
{
    uint8_t Timer4msOV      : 1;
    uint8_t Timer10msOV     : 1;
    uint8_t Timer50msOV     : 1;
    uint8_t Timer100msOV    : 1;
    uint8_t TimersActive    : 1;
} TimerFlags;

static volatile uint32_t Ticker;
static uint8_t TimerState;

static uint8_t TimeBase4ms;
static uint8_t TimeBase10ms;
static uint8_t TimeBase100ms;

void InitTimer(void)
{
    uint8_t sreg = SREG;
    cli();

    // Timer 0 initialisieren, Periode 100us, Vorteiler 8, CTC Mode
#if defined(__AVR_ATmega32__)
    OCR0 = (F_CPU / 80000UL) - 1;
    TCNT0 = 0;
    TCCR0 = (1<<WGM01)|(1<<CS01);
    TIMSK = (1<<OCIE0);
#elif defined(__AVR_ATmega324P__) || defined(__AVR_ATmega644__) || defined(__AVR_ATmega644P__)
    OCR0A = (F_CPU / 80000UL) - 1;
    OCR0B = 0;
    TCNT0 = 0;
    TCCR0A = (1<<WGM01);;
    TCCR0B = (1<<CS01);
    TIMSK0 = (1<<OCIE0A);
#else
#error Please define your TIMER0 code
#endif

    // Timer 2 initialisieren, Periode 500us, Vorteiler 32, CTC Mode
#if defined(__AVR_ATmega32__)
    OCR2 = (F_CPU / 64000UL) - 1;
    TCNT2 = 0;
    TCCR2 = (1<<WGM21)|(1<<CS21)|(1<<CS20);
    TIMSK |= (1<<OCF2);
#elif defined(__AVR_ATmega324P__) || defined(__AVR_ATmega644__)  || defined(__AVR_ATmega644P__)
    OCR2A = (F_CPU / 64000UL) - 1;
    OCR2B = 0;
    TCNT2 = 0;
    TCCR2A = (1<<WGM21);
    TCCR2B = (1<<CS21)|(1<<CS20);
    TIMSK2 |= (1<<OCIE2A);
#else
#error Please define your TIMER2 code
#endif


    TimerFlags.TimersActive = 0;

    SREG = sreg;
}

#if defined(__AVR_ATmega32__)
ISR(TIMER0_COMP_vect)
#elif defined(__AVR_ATmega324P__) || defined(__AVR_ATmega644__) || defined(__AVR_ATmega644P__)
ISR(TIMER0_COMPA_vect)
#else
#error Please define your TIMER0 code
#endif
{
    Ticker++;
}

#if defined(__AVR_ATmega32__)
ISR(TIMER2_COMP_vect)
#elif defined(__AVR_ATmega324P__) || defined(__AVR_ATmega644__) || defined(__AVR_ATmega644P__)
ISR(TIMER2_COMPA_vect)
#else
#error Please define your TIMER2 code
#endif
{
    static uint8_t lastRangeU = 0xff;
    static uint8_t stateRangeU = 0xff;
    static uint8_t lastRangeI = 0xff;
    static uint8_t stateRangeI = 0xff;
	static uint16_t lastUDACOut = 0xffff;
	static uint16_t lastIDACOut = 0xffff;
	uint8_t i;
    
	typedef union
    {
        uint32_t u32;
        uint16_t u16;
    } Values;
	
	#ifdef DUAL_DAC
	typedef struct
	{
		uint16_t U;
		uint16_t I;
	} DACData;
	
	static DACData DACOut = {
		.U = 0,
		.I = 0,
	};
	#endif
	
	static Values Value;
	

    // Interrupts freigeben
#if defined(__AVR_ATmega32__)
    TIMSK &= ~(1<<OCF2);
#elif defined(__AVR_ATmega324P__) || defined(__AVR_ATmega644__) || defined(__AVR_ATmega644P__)
    TIMSK2 &= ~(1<<OCIE2A);
#else
#error Please define your TIMER2 code
#endif
    sei();

    if (TimerState & 0x01)
    {
		#ifndef DUAL_DAC
        // MUX fr DAC abschalten
        PORTC &= ~(1<<PC4);     // MPXU abschalten, 0=off
        PORTC |= (1<<PC5);      // MPXI abschalten, 1=off
		
	
        if (TimerState & 0x02)
        {
		#endif
            // 3 -> Spannung ausgeben
            if (lastRangeU != RangeU)
            {
                // Spannungsbereich hat sich gendert
                stateRangeU = 0;
                lastRangeU = RangeU;
            }
            switch (stateRangeU)
            {
                default:
                    #ifdef DUAL_DAC
					DACOut.U = DACRawU;
					#else
					Value.u16 = DACRawU;
					#endif
                    break;

                case 0:
                case 1:
                    // fr 2 Zyklen (2x1ms) 0V ausgeben
                    #ifdef DUAL_DAC
					DACOut.U = 0;
					#else
					Value.u16 = 0;
					#endif
                    stateRangeU++;
                    break;

                case 2:
                case 3:
                    // fr 2 Zyklen (2x1ms) 0V ausgeben und Spannungsbereich umschalten
                    #ifdef DUAL_DAC
					DACOut.U = 0;
					#else
					Value.u16 = 0;
					#endif
                    if (lastRangeU)
                    {
                        PORTB |= (1<<PB5);
                    }
                    else
                    {
                        PORTB &= ~(1<<PB5);
                    }
                    stateRangeU++;
                    break;
            }
        #ifndef DUAL_DAC 
		}
        else
        {
		#endif
            // 1 -> Strom ausgeben
            if (lastRangeI != RangeI)
            {
                // Strombereich hat sich gendert
                stateRangeI = 0;
                lastRangeI = RangeI;
            }
            switch (stateRangeI)
            {
                default:
                    #ifdef DUAL_DAC
					DACOut.I = DACRawI;
					#else
					Value.u16 = DACRawI;
					#endif
					break;

                case 0:
                case 1:
                    // fr 2 Zyklen (2x1ms) 0A ausgeben
                    #ifdef DUAL_DAC
					DACOut.I = 0;
					#else
					Value.u16 = 0;
					#endif
					stateRangeI++;
                    break;

                case 2:
                case 3:
                    // fr 2 Zyklen (2x1ms) 0A ausgeben umd Strombereich umschalten
                    #ifdef DUAL_DAC
					DACOut.I = 0;
					#else
					Value.u16 = 0;
					#endif
                    PORTC = (PORTC & ~((1<<PC3)|(1<<PC2))) | (~(lastRangeI << 2) & ((1<<PC3)|(1<<PC2)));
                    stateRangeI++;
                    break;
            }
        #ifndef DUAL_DAC
		}
		#endif
		
		#ifdef DUAL_DAC
		if (lastUDACOut != DACOut.U)
		{
			lastUDACOut = DACOut.U;
			
			if (Params.Options.DAC16Present)
			{
				ShiftOut1655(DACOut.U, 0); //Lade den Spannungs-DAC
			}
			else
			{
				ShiftOut1257(DACOut.U);
			}
		}	
		if (lastIDACOut != DACOut.I)
		{
			lastIDACOut = DACOut.I;
			
			if (Params.Options.DAC16Present)
			{
				ShiftOut1655(DACOut.I, 1); //Lade den Strom-DAC
			}
			else
			{
				ShiftOut1257(DACOut.I);
			}
		}	
		
		#else
		
        if (Params.Options.DAC16Present)
        {
            ShiftOut1655(Value.u16);
        }
        else
        {
            ShiftOut1257(Value.u16);
        }
		#endif

        // Periodische Timer aktualisieren
        if (TimerFlags.TimersActive)
        {
            if (++TimeBase4ms >= 4)
            {
                TimerFlags.Timer4msOV = 1;
                TimeBase4ms = 0;
            }
            if (++TimeBase10ms >= 10)
            {
                TimerFlags.Timer10msOV = 1;
                TimeBase10ms = 0;
            }
            if (++TimeBase100ms >= 100)
            {
                TimerFlags.Timer100msOV = 1;
                TimeBase100ms = 0;
            }
            if (TimeBase100ms == 0 ||
                TimeBase100ms == 50)
            {
                TimerFlags.Timer50msOV = 1;
            }
        }
        jobEncoder();                   // runtime is part of the DAC settle time
        _delay_us(20);                  // more settle time

        #ifndef DUAL_DAC
		if (TimerState & 0x02)
        {
            // 3
            PORTC |= (1<<PC4);          // MUXU fr DAC einschalten, 1=on
        }
        else
        {
            // 1
            PORTC &= ~(1 << PC5);       // MPXI fr DACeinschalten, 0=on
        }
		#endif

    }
    else
    {
        // AD Wandlung
        if (Params.Options.ADC16Present)
        {
            PORTB |= (1<<PB0);          // SCLK high
            _delay_us(1);
            PORTB &= ~(1<<PB7);         // STRADC low
            Value.u32 = 0;
            _delay_us(4);
            PORTB |= (1<<PB7);          // STRADC high, Wandlung wird gestartet
            ShiftIn1864();
            for (i = 0; i < 4; i++)
            {
                Value.u32 += ShiftIn1864();
            }
            Value.u32 /= 4;

            if (TimerState & 0x02)
            {
                // 2
                ADCRawU = Value.u16;
                PORTC &= ~(1<<PC6);     // MUX fr ADC auf I
            }
            else
            {
                // 0,
                ADCRawI = Value.u16;
                PORTC |= (1<<PC6);      // MUX fr ADC auf U
            }
        }
        jobEncoder();
    }



    if (++TimerState >= 4)
    {
        TimerState = 0;
    }

    // Interrupts sperren
    cli();
#if defined(__AVR_ATmega32__)
    TIMSK |= (1<<OCF2);
#elif defined(__AVR_ATmega324P__) || defined(__AVR_ATmega644__) || defined(__AVR_ATmega644P__)
    TIMSK2 |= (1<<OCIE2A);
#else
#error Please define your TIMER2 code
#endif
}

void StartTimers(void)
{
    uint8_t sreg;

    sreg = SREG;
    cli();

    TimerFlags.Timer4msOV = 0;
    TimerFlags.Timer10msOV = 0;
    TimerFlags.Timer50msOV = 0;
    TimerFlags.Timer100msOV = 0;
    TimerFlags.TimersActive = 1;

    TimeBase4ms = 0;
    TimeBase10ms = 0;
    TimeBase100ms = 0;

    SREG = sreg;
}

uint8_t TestAndResetTimerOV(uint8_t TimerId)
{
    uint8_t result = 0;
    uint8_t sreg;

    switch(TimerId)
    {
        case TIMER_4MS:
            sreg = SREG;
            cli();
            if (TimerFlags.Timer4msOV)
            {
                TimerFlags.Timer4msOV = 0;
                result = 1;
            }
            SREG = sreg;
            break;

        case TIMER_10MS:
            sreg = SREG;
            cli();
            if (TimerFlags.Timer10msOV)
            {
                TimerFlags.Timer10msOV = 0;
                result = 1;
            }
            SREG = sreg;
            break;

        case TIMER_50MS:
            sreg = SREG;
            cli();
            if (TimerFlags.Timer50msOV)
            {
                TimerFlags.Timer50msOV = 0;
                result = 1;
            }
            SREG = sreg;
            break;

        case TIMER_100MS:
            sreg = SREG;
            cli();
            if (TimerFlags.Timer100msOV)
            {
                TimerFlags.Timer100msOV = 0;
                result = 1;
            }
            SREG = sreg;
            break;
    }
    return result;
}

uint32_t GetTicker(void)
{
    uint32_t result;
    uint8_t sreg;

    sreg = SREG;
    cli();

    result = Ticker;

    SREG = sreg;

    return result;
}

void wait_us(uint32_t us)
{
    uint32_t start, stop, current;

    current = start = GetTicker();

    stop = start + us / 100;

    if (stop < start)
    {
        while (current >= start)
        {
            current = GetTicker();
        }
    }
    while (current < stop)
    {
        current = GetTicker();
    }
}
